from abc import ABC, abstractmethod
from functools import wraps
from typing import List, Callable
import re

from rdkit import Chem
from rdkit.Chem import Mol


class Reward(ABC):
    @staticmethod
    @abstractmethod
    def get_objective_functions(conf: dict) -> List[Callable[[Mol], float]]:
        raise NotImplementedError('Please check your reward file')
    
    @staticmethod
    @abstractmethod
    def calc_reward_from_objective_values(values: List[float], conf: dict) -> float:
        raise NotImplementedError('Please check your reward file')


class BatchReward(ABC):
    @staticmethod
    @abstractmethod
    def get_batch_objective_functions() -> List[Callable[[List[Mol], List[dict]], float]]:
        raise NotImplementedError('Please check your reward file')
    
    @staticmethod
    @abstractmethod
    def calc_reward_from_objective_values(values: List[float], conf: dict) -> float:
        raise NotImplementedError('Please check your reward file')


def convert_to_linker_reward(conf: dict):
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            if not isinstance(args[0], Mol):
                raise TypeError("Check this decorator is placed in the correct position.")
            if 'cores' not in conf:
                raise KeyError("Must specify SMILES strings corresponding to the key `cores` in the config file.")
            smi = Chem.MolToSmiles(args[0])
            if smi.count("*") != len(conf['cores']):
                return -1
            mol_ = Chem.MolFromSmiles(add_atom_index_in_wildcard(smi))
            rwmol = Chem.RWMol(mol_)
            cores_mol = [Chem.MolFromSmiles(s) for s in conf['cores']]
            for m in cores_mol:
                rwmol.InsertMol(m)
            try:
                prod = Chem.molzip(rwmol)
            except:
                return -1
            return func(prod) 
        return wrapper

    def add_atom_index_in_wildcard(smiles: str):
        c = iter(range(1, smiles.count('*')+1))
        labeled_smiles = re.sub(r'\*', lambda _: f'[*:{next(c)}]', smiles)
        return labeled_smiles
    
    return decorator


import argparse
import ast
import os

import pandas as pd


def get_parser():
    parser = argparse.ArgumentParser(
        description="",
        usage=f"chemtsv2-column-splitter -i INPUT_FILE -t TARGET_COL_NAME -n NEW_COL_NAMES01 NEW_COL_NAMES02 ..."
    )
    parser.add_argument(
        "-i", "--input_file", type=str, required=True,
        help="path to a result CSV file"
    )
    parser.add_argument(
        "-t", "--target_column", type=str, required=True,
        help="target column name to be split"
    )
    parser.add_argument(
        "-n", "--new_column_names", type=str, nargs="*",required=True,
        help="new column names to be created from the target column"
    )
    return parser.parse_args()

    
def main():
    args = get_parser()
    print(f"[INFO] Input file: {args.input_file}\n"
          f"[INFO] Target column name: {args.target_column}\n"
          f"[INFO] New column names: {args.new_column_names}\n")

    df = pd.read_csv(args.input_file)
    df[args.target_column] = df[args.target_column].apply(ast.literal_eval)
    df[args.new_column_names] = pd.DataFrame(df[args.target_column].tolist(), index=df.index)
    df.drop(args.target_column, axis=1, inplace=True)

    stem, ext = os.path.splitext(args.input_file)
    output_fname = f"{stem}_column_{args.target_column}_split{ext}"
    df.to_csv(output_fname, mode='w', index=False)
    print(f"[INFO] Save to {output_fname}\n"
          f"[INFO] Done!")


if __name__ == "__main__":
    main()



from collections import deque
from copy import deepcopy
import datetime
from enum import Enum
from math import log, sqrt
import os
import pickle
import random  # only for Hash table initialization
import sys
import time

from mpi4py import MPI
import numpy as np
import pandas as pd
from rdkit import Chem

from chemtsv2.mp_utils import (
    backtrack_tdsdfuct, backtrack_mpmcts, compare_ucb_tdsdfuct, compare_ucb_mpmcts, update_selection_ucbtable_mpmcts, update_selection_ucbtable_tdsdfuct,
    Item, HashTable)
from chemtsv2.utils import chem_kn_simulation, build_smiles_from_tokens, expanded_node, has_passed_through_filters

"""
classes defined distributed parallel mcts
"""
class JobType(Enum):
    '''
    defines JobType tag values
    values higher than PRIORITY_BORDER (128) mean high prority tags
    FINISH is not used in this implementation. It will be needed for games.
    '''
    SEARCH = 0
    BACKPROPAGATION = 1
    PRIORITY_BORDER = 128
    CHECKPOINT_PREPARE = 150
    CHECKPOINT_READY = 151
    CHECKPOINT_SAVE = 152
    CHECKPOINT_LOAD = 153
    GATHER_RESULTS = 253
    TIMEUP = 254
    FINISH = 255

    @classmethod
    def is_high_priority(self, tag):
        return tag >= self.PRIORITY_BORDER.value


class Tree_Node():
    def __init__(self, state, parentNode=None, reward_calculator=None, conf=None):
        # todo: these should be in a numpy array
        # MPI payload [node.state, node.reward, node.wins, node.visits, node.num_thread_visited, node.path_ucb]
        self.state = state
        self.childNodes = []
        self.parentNode = parentNode
        self.wins = 0
        self.visits = 0
        self.virtual_loss = 0
        self.num_thread_visited = 0
        self.reward = 0
        self.check_childnode = []
        self.expanded_nodes = []
        self.path_ucb = []
        self.childucb = []
        self.conf = conf
        self.reward_calculator = reward_calculator
        self.val = conf['token']
        self.max_len=conf['max_len']

    def selection(self):
        ucb = []
        for i in range(len(self.childNodes)):
            ucb.append((self.childNodes[i].wins +
                        self.childNodes[i].virtual_loss) /
                       (self.childNodes[i].visits +
                        self.childNodes[i].num_thread_visited) +
                       self.conf['c_val'] *
                       sqrt(2 *log(self.visits +self.num_thread_visited) /
                            (self.childNodes[i].visits +
                                self.childNodes[i].num_thread_visited)))
        self.childucb = ucb
        m = np.amax(ucb)
        indices = np.nonzero(ucb == m)[0]
        ind = self.conf['random_generator'].choice(indices)
        self.childNodes[ind].num_thread_visited += 1
        self.num_thread_visited += 1
        return ind, self.childNodes[ind]

    def expansion(self, model, logger):
        node_idxs = expanded_node(model, self.state, self.val, logger)
        self.check_childnode.extend(node_idxs)
        self.expanded_nodes.extend(node_idxs)

    def addnode(self, m):
        self.expanded_nodes.remove(m)
        added_nodes = []
        added_nodes.extend(self.state)
        added_nodes.append(self.val[m])
        self.num_thread_visited += 1
        n = Tree_Node(state=added_nodes, parentNode=self, conf=self.conf)
        n.num_thread_visited += 1
        self.childNodes.append(n)
        return n

    def update_local_node(self, score):
        self.visits += 1
        self.wins += score
        self.reward = score

    def simulation(self, chem_model, state, gen_id, generated_dict):
        filter_flag = 0

        self.conf['gid'] = gen_id
        all_posible = chem_kn_simulation(chem_model, state, self.val, self.conf)
        smi = build_smiles_from_tokens(all_posible, self.val, use_selfies=self.conf['use_selfies'])

        if smi in generated_dict:
            values_list = generated_dict[smi][0]
            score = generated_dict[smi][1]
            filter_flag = generated_dict[smi][2]
            valid_flag = 1  # because only valid SMILES strings are stored in generated_dict
            return values_list, score, smi, filter_flag, valid_flag

        if has_passed_through_filters(smi, self.conf):
            mol = Chem.MolFromSmiles(smi)
            values_list = [f(mol) for f in self.reward_calculator.get_objective_functions(self.conf)]
            score = self.reward_calculator.calc_reward_from_objective_values(values=values_list, conf=self.conf)
            filter_flag = 1
            valid_flag = 1
        else:
            mol = Chem.MolFromSmiles(smi)
            if mol is None:
                values_list = [-999 for _ in self.reward_calculator.get_objective_functions(self.conf)]
                valid_flag = 0
            else:
                values_list = [f(mol) for f in self.reward_calculator.get_objective_functions(self.conf)]
                valid_flag = 1
            score = 0
            filter_flag = 0
        if valid_flag:
            generated_dict[smi] = [values_list, score, filter_flag]
        return values_list, score, smi, filter_flag, valid_flag

    def backpropagation(self, cnode):
        self.wins += cnode.reward
        self.visits += 1
        self.num_thread_visited -= 1
        self.reward = cnode.reward
        for i in range(len(self.childNodes)):
            if cnode.state[-1] == self.childNodes[i].state[-1]:
                self.childNodes[i].wins += cnode.reward
                self.childNodes[i].num_thread_visited -= 1
                self.childNodes[i].visits += 1


class p_mcts:
    """
    parallel mcts algorithms includes TDS-UCT, TDS-df-UCT and MP-MCTS
    """
    # todo: use generated_dict

    def __init__(self, communicator, chem_model, reward_calculator, conf, logger):
        self.comm = communicator
        self.rank = self.comm.Get_rank()
        self.nprocs = self.comm.Get_size()

        self.chem_model = chem_model
        self.reward_calculator = reward_calculator
        self.conf = conf
        self.logger = logger
        self.threshold = 3600 * conf['hours']
        # Initialize HashTable
        root_node = Tree_Node(state=['&'], reward_calculator=reward_calculator, conf=conf)
        random.seed(conf['zobrist_hash_seed'])
        self.hsm = HashTable(self.nprocs, root_node.val, root_node.max_len, len(root_node.val))

        if self.conf['checkpoint_load']:
            dt = datetime.datetime.now()
            stime = dt.strftime("%Y%m%d_%H%M%S")
            self.output_path = os.path.join(conf['output_dir'], f"result_C{conf['c_val']}_{stime}.csv")
        else:
            self.output_path = os.path.join(conf['output_dir'], f"result_C{conf['c_val']}.csv")
        if self.rank == 0:
            if os.path.exists(self.output_path):
                print(f"[ERROR] {self.output_path} already exists. Please specify a different file name.", file=sys.stderr, flush=True)
                self.comm.Abort()

        self.start_time = time.time()

        self.id_suffix = "_" + str(self.rank).zfill(len(str(self.nprocs)))

        # for results
        self.total_valid_num = 0

        # these are gathred to rank 0 in the end
        self.generated_id_list = []
        self.valid_smiles_list = []
        self.depth_list = []
        self.objective_values_list = [] # raw reward (could be list)
        self.reward_values_list = [] # normalized reward for UCT
        self.elapsed_time_list = []

        self.generated_dict = {}  # dictionary of generated compounds
        self.filter_check_list = [] # only needed for output (at rank 0)
        self.obj_column_names = [f.__name__ for f in self.reward_calculator.get_objective_functions(self.conf)]

    def get_generated_id(self):
        _id = str(self.total_valid_num) + self.id_suffix
        self.total_valid_num += 1
        return _id

    def elapsed_time(self):
        return time.time() - self.start_time

    def send_message(self, node, dest, tag, data=None):
        # send node using MPI_Bsend
        # typical usage of data is path_ucb for newly created child nodes
        if data is None:
            self.comm.bsend(np.asarray(
                [node.state, node.reward, node.wins, node.visits, node.num_thread_visited, node.path_ucb], dtype=object),
                dest=dest, tag=tag)
        else:
            self.comm.bsend(np.asarray(
                [node.state, node.reward, node.wins, node.visits, node.num_thread_visited, data], dtype=object),
                dest=dest, tag=tag)

    def send_search_childnode(self, node, ucb_table, dest):
        self.comm.bsend(np.asarray(
            [node.state, node.reward, node.wins, node.visits, node.num_thread_visited, ucb_table], dtype=object),
            dest=dest, tag=JobType.SEARCH.value)

    def send_backprop(self, node, dest):
        self.comm.bsend(np.asarray(
            [node.state, node.reward, node.wins, node.visits, node.num_thread_visited, node.path_ucb], dtype=object),
            dest=dest, tag=JobType.BACKPROPAGATION.value)

    def record_result(self, smiles, depth, reward, gen_id, raw_reward_list, filter_flag):
        self.valid_smiles_list.append(smiles)
        self.depth_list.append(depth)
        self.reward_values_list.append(reward)
        self.elapsed_time_list.append(self.elapsed_time())
        self.generated_id_list.append(gen_id)
        self.objective_values_list.append(raw_reward_list)
        self.filter_check_list.append(filter_flag)

    def gather_results(self):
        status = MPI.Status()
        if self.rank == 0:
            self.logger.info(f"Gather each rank result...")
        for id in range(1, self.nprocs):
            if self.rank == 0:
                (valid_smiles_list, depth_list, reward_values_list, elapsed_time_list,
                 generated_id_list, objective_values_list, filter_check_list) = self.comm.recv(source=id, tag=JobType.GATHER_RESULTS.value, status=status)
                self.valid_smiles_list.extend(valid_smiles_list)
                self.depth_list.extend(depth_list)
                self.reward_values_list.extend(reward_values_list)
                self.elapsed_time_list.extend(elapsed_time_list)
                self.generated_id_list.extend(generated_id_list)
                self.objective_values_list.extend(objective_values_list)
                self.filter_check_list.extend(filter_check_list)
            elif self.rank == id:
                self.comm.send((self.valid_smiles_list, self.depth_list, self.reward_values_list, self.elapsed_time_list,
                                self.generated_id_list, self.objective_values_list, self.filter_check_list),
                                dest=0, tag=JobType.GATHER_RESULTS.value)

    def flush(self):
        df = pd.DataFrame({
            "generated_id": self.generated_id_list,
            "smiles": self.valid_smiles_list,
            "reward": self.reward_values_list,
            "depth": self.depth_list,
            "elapsed_time": self.elapsed_time_list,
            "is_through_filter": self.filter_check_list,
        })
        df_obj = pd.DataFrame(self.objective_values_list, columns=self.obj_column_names)
        df = pd.concat([df, df_obj], axis=1)
        if os.path.exists(self.output_path):
            df.to_csv(self.output_path, mode='a', index=False, header=False)
        else:
            df.to_csv(self.output_path, mode='w', index=False)
        if self.rank == 0:
            self.logger.info(f"Save a result at {self.output_path}")

        self.generated_id_list.clear()
        self.valid_smiles_list.clear()
        self.reward_values_list.clear()
        self.depth_list.clear()
        self.elapsed_time_list.clear()
        self.filter_check_list.clear()
        self.objective_values_list.clear()
    
    def TDS_UCT(self):
        # self.comm.barrier()
        status = MPI.Status()

        self.start_time = time.time()
        _, rootdest = self.hsm.hashing(['&'])
        jobq = deque()
        timeup = False
        if self.rank == rootdest:
            root_job_message = np.asarray([['&'], None, 0, 0, 0, []], dtype=object)
            for i in range(3 * self.nprocs):
                temp = deepcopy(root_job_message)
                root_job = (JobType.SEARCH.value, temp)
                jobq.appendleft(root_job)
        while not timeup:
            if self.rank == 0:
                if self.elapsed_time() > self.threshold:
                    timeup = True
                    for dest in range(1, self.nprocs):
                        dummy_data = tag = JobType.TIMEUP.value
                        self.comm.bsend(dummy_data, dest=dest,
                                        tag=JobType.TIMEUP.value)
            while True:
                ret = self.comm.Iprobe(source=MPI.ANY_SOURCE,
                                        tag=MPI.ANY_TAG, status=status)
                if ret == False:
                    break
                else:
                    message = self.comm.recv(
                        source=MPI.ANY_SOURCE, tag=MPI.ANY_TAG, status=status)
                    cur_status = status
                    tag = cur_status.Get_tag()
                    job = (tag, message)
                    if JobType.is_high_priority(tag):
                        # high priority messages (timeup and finish)
                        jobq.append(job)
                    else:
                        # normal messages (search and backpropagate)
                        jobq.appendleft(job)

            jobq_non_empty = bool(jobq)
            if jobq_non_empty:
                (tag, message) = jobq.pop()
                if tag == JobType.SEARCH.value:
                    # if node is not in the hash table
                    if self.hsm.search_table(message[0]) is None:
                        node = Tree_Node(state=message[0], reward_calculator=self.reward_calculator, conf=self.conf)
                        #node.state = message[0]
                        if node.state == ['&']:
                            node.expansion(self.chem_model, self.logger)
                            m = self.conf['random_generator'].choice(node.expanded_nodes)
                            n = node.addnode(m)
                            self.hsm.insert(Item(node.state, node))
                            _, dest = self.hsm.hashing(n.state)
                            self.send_message(n, dest, tag=JobType.SEARCH.value)
                        else:
                            # or max_len_wavelength :
                            if len(node.state) < node.max_len:
                                gen_id = self.get_generated_id()
                                values_list, score, smi, filter_flag, is_valid_smi = node.simulation(
                                    self.chem_model, node.state, gen_id, self.generated_dict)
                                if is_valid_smi:
                                    self.record_result(smiles=smi, depth=len(node.state), reward=score,
                                                       gen_id=gen_id, raw_reward_list=values_list, filter_flag=filter_flag)
                                # backpropagation on local memory
                                node.update_local_node(score)
                                self.hsm.insert(Item(node.state, node))
                                _, dest = self.hsm.hashing(node.state[0:-1])
                                self.send_backprop(node, dest)
                            else:
                                score = -1
                                # backpropagation on local memory
                                node.update_local_node(score)
                                self.hsm.insert(Item(node.state, node))
                                _, dest = self.hsm.hashing(node.state[0:-1])
                                self.send_backprop(node, dest)

                    else:  # if node already in the local hashtable
                        node = self.hsm.search_table(message[0])
                        #print("debug:", node.visits,
                        #      node.num_thread_visited, node.wins)
                        if node.state == ['&']:
                            if node.expanded_nodes != []:
                                m = self.conf['random_generator'].choice(node.expanded_nodes)
                                n = node.addnode(m)
                                self.hsm.insert(Item(node.state, node))
                                _, dest = self.hsm.hashing(n.state)
                                self.comm.bsend(np.asarray([n.state, n.reward, n.wins, n.visits,
                                                n.num_thread_visited], dtype=object), dest=dest, tag=JobType.SEARCH.value)
                            else:
                                ind, childnode = node.selection()
                                self.hsm.insert(Item(node.state, node))
                                _, dest = self.hsm.hashing(childnode.state)
                                self.send_message(childnode, dest, tag=JobType.SEARCH.value)
                        else:
                            #node.num_thread_visited = message[4]
                            if len(node.state) < node.max_len:
                                if node.state[-1] != '\n':
                                    if node.expanded_nodes != []:
                                        m = self.conf['random_generator'].choice(node.expanded_nodes)
                                        n = node.addnode(m)
                                        self.hsm.insert(Item(node.state, node))
                                        _, dest = self.hsm.hashing(n.state)
                                        self.send_message(n, dest, tag=JobType.SEARCH.value)
                                    else:
                                        if node.check_childnode == []:
                                            node.expansion(self.chem_model, self.logger)
                                            m = self.conf['random_generator'].choice(
                                                node.expanded_nodes)
                                            n = node.addnode(m)
                                            self.hsm.insert(Item(node.state, node))
                                            _, dest = self.hsm.hashing(n.state)
                                            self.send_message(n, dest, tag=JobType.SEARCH.value)
                                        else:
                                            ind, childnode = node.selection()
                                            self.hsm.insert(Item(node.state, node))
                                            _, dest = self.hsm.hashing(
                                                childnode.state)
                                            self.send_message(childnode, dest, tag=JobType.SEARCH.value)

                                else:
                                    gen_id = self.get_generated_id()
                                    values_list, score, smi, filter_flag, is_valid_smi = node.simulation(
                                        self.chem_model, node.state, gen_id, self.generated_dict)
                                    score = -1
                                    if is_valid_smi:
                                        self.record_result(smiles=smi, depth=len(node.state), reward=score,
                                                           gen_id=gen_id, raw_reward_list=values_list, filter_flag=filter_flag)
                                    # backpropagation on local memory
                                    node.update_local_node(score)
                                    self.hsm.insert(Item(node.state, node))
                                    _, dest = self.hsm.hashing(node.state[0:-1])
                                    self.send_backprop(node, dest)

                            else:
                                score = -1
                                # backpropagation on local memory
                                node.update_local_node(score)
                                self.hsm.insert(Item(node.state, node))
                                _, dest = self.hsm.hashing(node.state[0:-1])
                                self.send_backprop(node, dest)

                elif tag == JobType.BACKPROPAGATION.value:
                    node = Tree_Node(state=message[0], reward_calculator=self.reward_calculator, conf=self.conf)
                    node.reward = message[1]
                    local_node = self.hsm.search_table(message[0][0:-1])
                    if local_node.state == ['&']:
                        local_node.backpropagation(node)
                        self.hsm.insert(Item(local_node.state, local_node))
                        _, dest = self.hsm.hashing(local_node.state)
                        self.send_message(local_node, dest, tag=JobType.SEARCH.value)
                    else:
                        local_node.backpropagation(node)
                        self.hsm.insert(Item(local_node.state, local_node))
                        _, dest = self.hsm.hashing(local_node.state[0:-1])
                        self.send_backprop(local_node, dest)
                elif tag == JobType.TIMEUP.value:
                    timeup = True

        return

    def TDS_df_UCT(self):
        # self.comm.barrier()
        status = MPI.Status()
        self.start_time = time.time()
        bpm = 0
        bp = []
        _, rootdest = self.hsm.hashing(['&'])
        jobq = deque()
        timeup = False
        if self.rank == rootdest:
            root_job_message = np.asarray([['&'], None, 0, 0, 0, []], dtype=object)
            for i in range(3 * self.nprocs):
                temp = deepcopy(root_job_message)
                root_job = (JobType.SEARCH.value, temp)
                jobq.appendleft(root_job)
        while not timeup:
            if self.rank == 0:
                if self.elapsed_time() > self.threshold:
                    timeup = True
                    for dest in range(1, self.nprocs):
                        dummy_data = tag = JobType.TIMEUP.value
                        self.comm.bsend(dummy_data, dest=dest,
                                        tag=JobType.TIMEUP.value)
            while True:
                ret = self.comm.Iprobe(source=MPI.ANY_SOURCE,
                                        tag=MPI.ANY_TAG, status=status)
                if ret == False:
                    break
                else:
                    message = self.comm.recv(
                        source=MPI.ANY_SOURCE, tag=MPI.ANY_TAG, status=status)
                    cur_status = status
                    tag = cur_status.Get_tag()
                    job = (tag, message)
                    if JobType.is_high_priority(tag):
                        jobq.append(job)
                    else:
                        jobq.appendleft(job)
            jobq_non_empty = bool(jobq)
            if jobq_non_empty:
                (tag, message) = jobq.pop()
                if tag == JobType.SEARCH.value:
                    if self.hsm.search_table(message[0]) == None:
                        node = Tree_Node(state=message[0], reward_calculator=self.reward_calculator, conf=self.conf)
                        info_table = message[5]
                        #print ("not in table info_table:",info_table)
                        if node.state == ['&']:
                            node.expansion(self.chem_model, self.logger)
                            m = self.conf['random_generator'].choice(node.expanded_nodes)
                            n = node.addnode(m)
                            self.hsm.insert(Item(node.state, node))
                            _, dest = self.hsm.hashing(n.state)
                            self.send_message(n, dest, tag=JobType.SEARCH.value)
                        else:
                            if len(node.state) < node.max_len:
                                gen_id = self.get_generated_id()
                                values_list, score, smi, filter_flag, is_valid_smi = node.simulation(
                                    self.chem_model, node.state, gen_id, self.generated_dict)
                                if is_valid_smi:
                                    self.record_result(smiles=smi, depth=len(node.state), reward=score,
                                                       gen_id=gen_id, raw_reward_list=values_list, filter_flag=filter_flag)
                                node.update_local_node(score)
                                # update infor table
                                info_table = backtrack_tdsdfuct(
                                    info_table, score)
                                self.hsm.insert(Item(node.state, node))
                                _, dest = self.hsm.hashing(node.state[0:-1])
                                self.send_backprop(node, dest)
                            else:
                                score = -1
                                node.update_local_node(score)
                                info_table = backtrack_tdsdfuct(
                                    info_table, score)
                                self.hsm.insert(Item(node.state, node))
                                _, dest = self.hsm.hashing(node.state[0:-1])
                                self.send_backprop(node, dest)

                    else:  # if node already in the local hashtable
                        node = self.hsm.search_table(message[0])
                        info_table = message[5]
                        #print ("in table info_table:",info_table)
                        if node.state == ['&']:
                            # print ("in table root:",node.state,node.path_ucb,len(node.state),len(node.path_ucb))
                            if node.expanded_nodes != []:
                                m = self.conf['random_generator'].choice(node.expanded_nodes)
                                n = node.addnode(m)
                                self.hsm.insert(Item(node.state, node))
                                _, dest = self.hsm.hashing(n.state)
                                self.send_message(n, dest, tag=JobType.SEARCH.value)
                            else:
                                ind, childnode = node.selection()
                                self.hsm.insert(Item(node.state, node))
                                info_table = update_selection_ucbtable_tdsdfuct(
                                    info_table, node, ind)
                                #print ("info_table after selection:",info_table)
                                _, dest = self.hsm.hashing(childnode.state)
                                self.send_message(childnode, dest, tag=JobType.SEARCH.value)
                        else:
                            #node.path_ucb = message[5]
                            # info_table=message[5]
                            #print("check ucb:", node.reward, node.visits, node.num_thread_visited,info_table)
                            if len(node.state) < node.max_len:
                                if node.state[-1] != '\n':
                                    if node.expanded_nodes != []:
                                        m = self.conf['random_generator'].choice(node.expanded_nodes)
                                        n = node.addnode(m)
                                        self.hsm.insert(Item(node.state, node))
                                        _, dest = self.hsm.hashing(n.state)
                                        self.send_message(n, dest, tag=JobType.SEARCH.value)
                                    else:
                                        if node.check_childnode == []:
                                            node.expansion(self.chem_model, self.logger)
                                            m = self.conf['random_generator'].choice(
                                                node.expanded_nodes)
                                            n = node.addnode(m)
                                            self.hsm.insert(Item(node.state, node))
                                            _, dest = self.hsm.hashing(n.state)
                                            self.send_message(n, dest, tag=JobType.SEARCH.value)
                                        else:
                                            ind, childnode = node.selection()
                                            self.hsm.insert(Item(node.state, node))
                                            info_table = update_selection_ucbtable_tdsdfuct(
                                                info_table, node, ind)
                                            _, dest = self.hsm.hashing(
                                                childnode.state)
                                            self.send_message(childnode, dest, tag=JobType.SEARCH.value)
                                else:
                                    gen_id = self.get_generated_id()
                                    value_list, score, smi, filter_flag, is_valid_smi = node.simulation(
                                        self.chem_model, node.state, gen_id, self.generated_dict)
                                    score = -1
                                    if is_valid_smi:
                                        self.record_result(smiles=smi, depth=len(node.state), reward=score,
                                                           gen_id=gen_id, raw_reward_list=value_list, filter_flag=filter_flag)
                                    node.update_local_node(score)
                                    info_table = backtrack_tdsdfuct(
                                        info_table, score)

                                    self.hsm.insert(Item(node.state, node))
                                    _, dest = self.hsm.hashing(node.state[0:-1])
                                    self.send_backprop(node, dest)
                            else:
                                score = -1
                                node.update_local_node(score)
                                info_table = backtrack_tdsdfuct(
                                    info_table, score)
                                self.hsm.insert(Item(node.state, node))
                                _, dest = self.hsm.hashing(node.state[0:-1])
                                self.send_backprop(node, dest)

                elif tag == JobType.BACKPROPAGATION.value:
                    bpm += 1
                    node = Tree_Node(state=message[0], reward_calculator=self.reward_calculator, conf=self.conf)
                    node.reward = message[1]
                    local_node = self.hsm.search_table(message[0][0:-1])
                    #print ("report check message[5]:",message[5])
                    #print ("check:",len(message[0]), len(message[5]))
                    #print ("check:",local_node.wins, local_node.visits, local_node.num_thread_visited)
                    info_table=message[5]
                    if local_node.state == ['&']:
                        local_node.backpropagation(node)
                        self.hsm.insert(Item(local_node.state, local_node))
                        _, dest = self.hsm.hashing(local_node.state)
                        self.send_message(local_node, dest, tag=JobType.SEARCH.value)
                    else:
                        local_node.backpropagation(node)
                        #local_node,info_table = backtrack_tdsdf(info_table,local_node, node)
                        back_flag = compare_ucb_tdsdfuct(info_table,local_node)
                        self.hsm.insert(Item(local_node.state, local_node))
                        if back_flag == 1:
                            _, dest = self.hsm.hashing(local_node.state[0:-1])
                            self.send_backprop(local_node, dest)
                        if back_flag == 0:
                            _, dest = self.hsm.hashing(local_node.state)
                            self.send_message(local_node, dest, tag=JobType.SEARCH.value)
                elif tag == JobType.TIMEUP.value:
                    timeup = True
        bp.append(bpm)

        return

    def MP_MCTS(self):
        #self.comm.barrier()
        status = MPI.Status()
        self.start_time = time.time()
        _, rootdest = self.hsm.hashing(['&'])
        jobq = deque()
        timeup = False
        checkpoint_prepare = False
        checkpoint_saved = False
        checkpoint_ready_count = 0
        if self.conf['checkpoint_load']:
            for i in range(self.nprocs):
                if self.rank == i:
                    self.comm.barrier()
                    ckpt_path = os.path.join(self.conf['output_dir'], f"mp_checkpoint_rank{i:04}.pickle")
                    with open(ckpt_path, mode='rb') as f:
                        cp_obj = pickle.load(f)
                    print('Load complete', self.rank)
            #self.generated_id_list = cp_obj['generated_id_list']
            self.total_valid_num = cp_obj['total_valid_num']
            #self.conf = cp_obj['conf']
            self.generated_dict = cp_obj['generated_dict']
            self.hsm = cp_obj['hsm']
            #self.start_time = cp_obj['start_time']
            jobq = cp_obj['jobq']
        elif self.rank == rootdest:
            root_job_message = np.asarray([['&'], None, 0, 0, 0, []], dtype=object)
            for i in range(3 * self.nprocs):
                temp = deepcopy(root_job_message)
                root_job = (JobType.SEARCH.value, temp)
                jobq.appendleft(root_job)

        while not timeup:
            if self.rank == 0:
                if self.elapsed_time() > self.threshold and self.conf['save_checkpoint'] and not checkpoint_prepare:
                    checkpoint_prepare = True
                    checkpoint_ready_count = 0
                    #print('Checkpoint prepare')
                    for dest in range(1, self.nprocs):
                        dummy_data = tag = JobType.CHECKPOINT_PREPARE.value
                        self.comm.bsend(dummy_data, dest=dest, tag=JobType.CHECKPOINT_PREPARE.value)
                if self.elapsed_time() > self.threshold and (not self.conf['save_checkpoint'] or checkpoint_saved):
                    timeup = True
                    for dest in range(1, self.nprocs):
                        dummy_data = tag = JobType.TIMEUP.value
                        self.comm.bsend(dummy_data, dest=dest, tag=JobType.TIMEUP.value)

            while True:
                ret = self.comm.Iprobe(source=MPI.ANY_SOURCE, tag=MPI.ANY_TAG, status=status)
                if ret == False:
                    break
                else:
                    message = self.comm.recv(source=MPI.ANY_SOURCE, tag=MPI.ANY_TAG, status=status)
                    cur_status = status
                    tag = cur_status.Get_tag()
                    job = (tag, message)
                    if JobType.is_high_priority(tag):
                        jobq.append(job)
                    else:
                        jobq.appendleft(job)
            jobq_non_empty = bool(jobq)
            if jobq_non_empty:
                if checkpoint_prepare:
                    (tag, message) = jobq[-1]
                    if JobType.is_high_priority(tag):
                        pass
                    else:
                        continue
                (tag, message) = jobq.pop()
                if tag == JobType.SEARCH.value:
                    if self.hsm.search_table(message[0]) == None:
                        node = Tree_Node(state=message[0], reward_calculator=self.reward_calculator, conf=self.conf)
                        if node.state == ['&']:
                            node.expansion(self.chem_model, self.logger)
                            m = self.conf['random_generator'].choice(node.expanded_nodes)
                            n = node.addnode(m)
                            self.hsm.insert(Item(node.state, node))
                            _, dest = self.hsm.hashing(n.state)
                            self.send_message(n, dest, tag=JobType.SEARCH.value)
                        else:
                            if len(node.state) < node.max_len:
                                gen_id = self.get_generated_id()
                                values_list, score, smi, filter_flag, is_valid_smi = node.simulation(
                                    self.chem_model, node.state, gen_id, self.generated_dict)
                                if is_valid_smi:
                                    self.record_result(smiles=smi, depth=len(node.state), reward=score,
                                                       gen_id=gen_id ,raw_reward_list=values_list, filter_flag=filter_flag)
                                node.update_local_node(score)
                                self.hsm.insert(Item(node.state, node))
                                _, dest = self.hsm.hashing(node.state[0:-1])
                                self.send_backprop(node, dest)
                            else:
                                score = -1
                                node.update_local_node(score)
                                self.hsm.insert(Item(node.state, node))
                                _, dest = self.hsm.hashing(node.state[0:-1])
                                self.send_backprop(node, dest)

                    else:  # if node already in the local hashtable
                        node = self.hsm.search_table(message[0])
                        if node.state == ['&']:
                            # print ("in table root:",node.state,node.path_ucb,len(node.state),len(node.path_ucb))
                            if node.expanded_nodes != []:
                                m = self.conf['random_generator'].choice(node.expanded_nodes)
                                n = node.addnode(m)
                                self.hsm.insert(Item(node.state, node))
                                _, dest = self.hsm.hashing(n.state)
                                self.send_message(n, dest, tag=JobType.SEARCH.value)
                            else:
                                ind, childnode = node.selection()
                                self.hsm.insert(Item(node.state, node))
                                ucb_table = update_selection_ucbtable_mpmcts(node, ind)
                                _, dest = self.hsm.hashing(childnode.state)
                                self.send_message(childnode, dest, tag=JobType.SEARCH.value, data=ucb_table)
                        else:
                            node.path_ucb = message[5]
                            #print("check ucb:", node.wins, node.visits, node.num_thread_visited)
                            if len(node.state) < node.max_len:
                                if node.state[-1] != '\n':
                                    if node.expanded_nodes != []:
                                        m = self.conf['random_generator'].choice(node.expanded_nodes)
                                        n = node.addnode(m)
                                        self.hsm.insert(Item(node.state, node))
                                        _, dest = self.hsm.hashing(n.state)
                                        self.send_message(n, dest, tag=JobType.SEARCH.value)
                                    else:
                                        if node.check_childnode == []:
                                            node.expansion(self.chem_model, self.logger)
                                            m = self.conf['random_generator'].choice(node.expanded_nodes)
                                            n = node.addnode(m)
                                            self.hsm.insert(Item(node.state, node))
                                            _, dest = self.hsm.hashing(n.state)
                                            self.send_message(n, dest, tag=JobType.SEARCH.value)
                                        else:
                                            ind, childnode = node.selection()
                                            self.hsm.insert(Item(node.state, node))
                                            ucb_table = update_selection_ucbtable_mpmcts(node, ind)
                                            _, dest = self.hsm.hashing(childnode.state)
                                            self.send_message(childnode, dest, tag=JobType.SEARCH.value, data=ucb_table)
                                else:
                                    gen_id = self.get_generated_id()
                                    values_list, score, smi, filter_flag, is_valid_smi = node.simulation(
                                        self.chem_model, node.state, gen_id, self.generated_dict)
                                    score = -1
                                    if is_valid_smi:
                                        self.record_result(smiles=smi, depth=len(node.state), reward=score,
                                                           gen_id=gen_id,raw_reward_list=values_list, filter_flag=filter_flag)
                                    node.update_local_node(score)
                                    self.hsm.insert(Item(node.state, node))
                                    _, dest = self.hsm.hashing(node.state[0:-1])
                                    self.send_backprop(node, dest)
                            else:
                                score = -1
                                node.update_local_node(score)
                                self.hsm.insert(Item(node.state, node))
                                _, dest = self.hsm.hashing(node.state[0:-1])
                                self.send_backprop(node, dest)

                elif tag == JobType.BACKPROPAGATION.value:
                    node = Tree_Node(state=message[0], reward_calculator=self.reward_calculator, conf=self.conf)
                    node.reward = message[1]
                    local_node = self.hsm.search_table(message[0][0:-1])
                    if local_node.state == ['&']:
                        local_node.backpropagation(node)
                        self.hsm.insert(Item(local_node.state, local_node))
                        _, dest = self.hsm.hashing(local_node.state)
                        self.send_message(local_node, dest, tag=JobType.SEARCH.value)
                    else:
                        local_node.backpropagation(node)
                        local_node = backtrack_mpmcts(local_node, node)
                        back_flag = compare_ucb_mpmcts(local_node)
                        self.hsm.insert(Item(local_node.state, local_node))
                        if back_flag == 1:
                            _, dest = self.hsm.hashing(local_node.state[0:-1])
                            self.send_backprop(local_node, dest)
                        if back_flag == 0:
                            _, dest = self.hsm.hashing(local_node.state)
                            self.send_message(local_node, dest, tag=JobType.SEARCH.value)

                elif tag == JobType.TIMEUP.value:
                    #print('Timeup', self.rank)
                    timeup = True
                elif tag == JobType.CHECKPOINT_PREPARE.value:
                    checkpoint_prepare = True
                    #print('Checkpoint prepare', self.rank)
                    dummy_data = JobType.CHECKPOINT_READY.value
                    self.comm.bsend(dummy_data, dest=0, tag=JobType.CHECKPOINT_READY.value)
                elif tag == JobType.CHECKPOINT_READY.value:
                    assert self.rank == 0
                    checkpoint_ready_count += 1
                    #print('Checkpoint ready count', checkpoint_ready_count)
                    if checkpoint_ready_count >= self.nprocs - 1:
                        #print('Checkpoint ready count', checkpoint_ready_count)
                        for dest in range(0, self.nprocs):
                            dummy_data = tag = JobType.CHECKPOINT_SAVE.value
                            self.comm.bsend(dummy_data, dest=dest, tag=JobType.CHECKPOINT_SAVE.value)
                elif tag == JobType.CHECKPOINT_SAVE.value:
                    cp_obj = {
                        #'generated_id_list': self.generated_id_list,
                        'total_valid_num': self.total_valid_num,
                        'conf': self.conf, 
                        'generated_dict': self.generated_dict,
                        'hsm': self.hsm,
                        'start_time': self.start_time,
                        'jobq': jobq,
                    }
                    #print('Checkpoint save start')
                    for i in range(self.nprocs):
                        if self.rank == i:
                            self.comm.barrier()
                            ckpt_path = os.path.join(self.conf['output_dir'], f"mp_checkpoint_rank{i:04}.pickle")
                            with open(ckpt_path, mode='wb') as f:
                                pickle.dump(cp_obj, f)
                    #print('Checkpoint save finished')
                    checkpoint_saved = True

        return



import os
import sys
import time

import numpy as np
import pandas as pd
import pickle

from chemtsv2.utils import chem_kn_simulation, build_smiles_from_tokens,\
    evaluate_node, node_to_add, expanded_node, back_propagation


class State:
    def __init__(self, position=['&'], parent=None):
        self.position = position
        self.visits = 0
        self.total_reward = 0
        self.parent_node = parent
        self.child_nodes = []
        
    def clone(self, include_visit=False, include_total_reward=False, include_parent_node=False, include_child_node=False):
        st = State()
        st.position = self.position[:]
        st.visits = self.visits if include_visit else 0
        st.total_reward = self.total_reward if include_total_reward else 0
        st.parent_node = self.parent_node if include_parent_node else None
        st.child_nodes = self.child_nodes if include_child_node else []
        return st

    def add_position(self, m):
        self.position.append(m)


class Node:
    def __init__(self, policy_evaluator, position=None, state=None, conf=None):
        self.position = position
        self.state = state
        self.policy_evaluator = policy_evaluator
        self.conf = conf

    def select_node(self, logger):
        score_list = []
        logger.debug('UCB:')
        for i in range(len(self.state.child_nodes)):
            score = self.policy_evaluator.evaluate(self.state.child_nodes[i].state, self.conf)
            score_list.append(score)
            logger.debug(f"{self.state.child_nodes[i].position} {score}") 
        m = np.amax(score_list)
        indices = np.nonzero(score_list == m)[0]
        ind = int(self.conf['random_generator'].choice(indices))
        s = self.state.child_nodes[ind]
        logger.debug(f"\nindex {ind} {self.position} {m}") 
        return s

    def add_node(self, m, state, policy_evaluator):
        state.parent_node = self
        node = Node(policy_evaluator, position=m, state=state, conf=self.conf)
        self.state.child_nodes.append(node)

    def simulation(self):
        raise SystemExit("[ERROR] Do NOT use this method")

    def update(self, reward):
        self.state.visits += 1
        self.state.total_reward += reward


class MCTS:
    def __init__(self, root_state, conf, tokens, model, reward_calculator, policy_evaluator, logger):
        self.start_time = time.time()
        self.rootnode = Node(policy_evaluator, state=root_state, conf=conf)
        self.conf = conf
        self.tokens = tokens
        self.model = model
        self.reward_calculator = reward_calculator
        self.policy_evaluator = policy_evaluator
        self.logger = logger

        self.valid_smiles_list = []
        self.depth_list = []
        self.objective_values_list = []
        self.reward_values_list = []
        self.elapsed_time_list = []
        self.generated_dict = {}  # dictionary of generated compounds
        self.generated_id_list = []
        self.filter_check_list = []
        self.total_valid_num = 0
        
        if conf['batch_reward_calculation']:
            self.obj_column_names = [f.__name__ for f in self.reward_calculator.get_batch_objective_functions()]
        else:
            self.obj_column_names = [f.__name__ for f in self.reward_calculator.get_objective_functions(self.conf)]
        self.output_path = os.path.join(conf['output_dir'], f"result_C{conf['c_val']}.csv")
        if os.path.exists(self.output_path) and not conf['restart']:
            sys.exit(f"[ERROR] {self.output_path} already exists. Please specify a different file name.")

        self.gid = 0
        self.loop_counter_for_selection = 0
        self.loop_counter_for_expansion = 0
        self.expanded_before = {}

        if conf['threshold_type'] == "time":
            self.threshold = time.time() + 3600 * conf['hours']
        elif conf['threshold_type'] == "generation_num":
            self.threshold = conf['generation_num']
        else:
            sys.exit("[ERROR] Specify 'threshold_type': [time, generation_num]")

    def flush(self):
        
        df = pd.DataFrame({
            "generated_id": self.generated_id_list,
            "smiles": self.valid_smiles_list,
            "reward": self.reward_values_list,
            "depth": self.depth_list,
            "elapsed_time": self.elapsed_time_list,
            "is_through_filter": self.filter_check_list,
        })
        df_obj = pd.DataFrame(self.objective_values_list, columns=self.obj_column_names)
        df = pd.concat([df, df_obj], axis=1)
        if os.path.exists(self.output_path):
            df.to_csv(self.output_path, mode='a', index=False, header=False)
        else:
            df.to_csv(self.output_path, mode='w', index=False)
        self.logger.info(f"save results at {self.output_path}")

        self.generated_id_list.clear()
        self.valid_smiles_list.clear()
        self.reward_values_list.clear()
        self.depth_list.clear()
        self.elapsed_time_list.clear()
        self.filter_check_list.clear()
        self.objective_values_list.clear()

    def search(self):
        """initialization of search tree"""
        ckpt_path = os.path.join(self.conf['output_dir'], self.conf['checkpoint_file'])
        if self.conf['restart'] and os.path.exists(ckpt_path):
            self.logger.info(f"Load the checkpoint file from {ckpt_path}")
            self.load_checkpoint()
        
        while (time.time() if self.conf['threshold_type']=="time" else self.total_valid_num) <= self.threshold:
            node = self.rootnode  # important! This node is different with state / node is the tree node
            state = node.state.clone()  # but this state is the state of the initialization. Too important!

            """selection step"""
            node_pool = []
            while node.state.child_nodes != []:
                node = node.select_node(self.logger)
                state.add_position(node.position)
            self.logger.info(f"state position: {state.position}")

            self.logger.debug(f"infinite loop counter (selection): {self.loop_counter_for_selection}")
            if node.position == '\n':
                back_propagation(node, reward=-1.0)
                self.loop_counter_for_selection += 1
                if self.loop_counter_for_selection > self.conf['infinite_loop_threshold_for_selection']:
                    self.flush()
                    sys.exit('[WARN] Infinite loop is detected in the selection step. Change hyperparameters or RNN model.')
                continue
            else:
                self.loop_counter_for_selection = 0

            """expansion step"""
            expanded = expanded_node(self.model, state.position, self.tokens, self.logger, threshold=self.conf['expansion_threshold'])
            self.logger.debug(f"infinite loop counter (expansion): {self.loop_counter_for_expansion}")
            if set(expanded) == self.expanded_before:
                self.loop_counter_for_expansion += 1
                if self.loop_counter_for_expansion > self.conf['infinite_loop_threshold_for_expansion']:
                    self.flush()
                    sys.exit('[WARN] Infinite loop is detected in the expansion step. Change hyperparameters or RNN model.')
            else:
                self.loop_counter_for_expansion = 0
            self.expanded_before = set(expanded)

            new_compound = []
            nodeadded = []
            for _ in range(self.conf['simulation_num']):
                nodeadded_tmp = node_to_add(expanded, self.tokens, self.logger)
                nodeadded.extend(nodeadded_tmp)
                for n in nodeadded_tmp:
                    position_tmp = state.position + [n]
                    all_posible = chem_kn_simulation(self.model, position_tmp, self.tokens, self.conf)
                    new_compound.append(build_smiles_from_tokens(all_posible, self.tokens, use_selfies=self.conf['use_selfies']))

            _gids = list(range(self.gid, self.gid+len(new_compound)))
            self.gid += len(new_compound)

            self.logger.debug(f"nodeadded {nodeadded}")
            self.logger.info(f"new compound {new_compound}")
            self.logger.debug(f"generated_dict {self.generated_dict}") 
            if self.conf["debug"]:
                self.logger.debug('\n' + '\n'.join([f"lastcomp {comp[-1]} ... " + str(comp[-1] == '\n') for comp in new_compound]))
            node_index, objective_values, valid_smiles, generated_id_list, filter_check_list = evaluate_node(new_compound, self.generated_dict, self.reward_calculator, self.conf, self.logger, _gids)

            if len(valid_smiles) == 0:
                back_propagation(node, reward=-1.0)
                continue

            valid_num = len(valid_smiles)
            self.total_valid_num += valid_num
            self.valid_smiles_list.extend(valid_smiles)
            depth = len(state.position)
            self.depth_list.extend([depth for _ in range(valid_num)])
            elapsed_time = round(time.time()-self.start_time, 1)
            self.elapsed_time_list.extend([elapsed_time for _ in range(valid_num)])
            self.objective_values_list.extend(objective_values)
            self.generated_id_list.extend(generated_id_list)
            self.filter_check_list.extend(filter_check_list)

            self.logger.info(f"Number of valid SMILES: {self.total_valid_num}")
            self.logger.debug(f"node {node_index} objective_values {objective_values} valid smiles {valid_smiles} time {elapsed_time}")

            re_list = []
            atom_checked = []
            for i in range(len(node_index)):
                m = node_index[i]
                atom = nodeadded[m]
                state_clone = state.clone(include_visit=True, include_total_reward=True)

                if atom not in atom_checked: 
                    node.add_node(atom, state_clone, self.policy_evaluator)
                    node_pool.append(node.state.child_nodes[len(atom_checked)])
                    atom_checked.append(atom)
                else:
                    node_pool.append(node.state.child_nodes[atom_checked.index(atom)])

                if self.conf["debug"]:
                    self.logger.debug('\n' + '\n'.join([f"Child node position ... {c.position}" for c in node.state.child_nodes]))

                re = -1 if atom == '\n' else self.reward_calculator.calc_reward_from_objective_values(values=objective_values[i], conf=self.conf)
                if self.conf['include_filter_result_in_reward']:
                    re *= filter_check_list[i]
                    
                re_list.append(re)
                self.logger.debug(f"atom: {atom} re_list: {re_list}")
            self.reward_values_list.extend(re_list)

            """backpropation step"""
            for i in range(len(node_pool)):
                node = node_pool[i]
                back_propagation(node, reward=re_list[i])

            if self.conf['debug']:
                self.logger.debug('\n' + '\n'.join([f"child position: {c.position}, total_reward: {c.state.total_reward}, visits: {c.state.visits}" for c in node_pool]))

            if len(self.valid_smiles_list) > self.conf['flush_threshold'] and self.conf['flush_threshold'] != -1:
                self.flush()
            
            """save checkpoint file"""
            if self.conf['save_checkpoint']:
                self.save_checkpoint()

        if len(self.valid_smiles_list) > 0:
            self.flush()
            
    def load_checkpoint(self):
        ckpt_path = os.path.join(self.conf['output_dir'], self.conf['checkpoint_file'])
        with open(ckpt_path, mode='rb') as f:
            cp_obj = pickle.load(f)
        self.gid = cp_obj['gid']
        self.loop_counter_for_selection = cp_obj['loop_counter_for_selection']
        self.loop_counter_for_expansion = cp_obj['loop_counter_for_expansion']
        self.expanded_before = cp_obj['expanded_before']        
        self.start_time = cp_obj['start_time']
        self.rootnode = cp_obj['rootnode']
        self.conf = cp_obj['conf']
        self.generated_dict = cp_obj['generated_dict']
        self.total_valid_num = cp_obj['total_valid_num']


    def save_checkpoint(self):
        ckpt_fname = self.conf['checkpoint_file']
        ckpt_path = os.path.join(self.conf['output_dir'], ckpt_fname)
        stem, ext = ckpt_fname.rsplit('.', 1)
        # To keep the three most recent checkpoint files.
        ckpt1_path = os.path.join(self.conf['output_dir'], f'{stem}2.{ext}')
        ckpt2_path = os.path.join(self.conf['output_dir'], f'{stem}3.{ext}')
        if os.path.exists(ckpt1_path):
            os.rename(ckpt1_path, ckpt2_path)
        if os.path.exists(ckpt_path):
            os.rename(ckpt_path, ckpt1_path)

        cp_obj = {
            'gid': self.gid,
            'loop_counter_for_selection': self.loop_counter_for_selection,
            'loop_counter_for_expansion': self.loop_counter_for_expansion,
            'expanded_before': self.expanded_before,
            'start_time': self.start_time, 
            'conf': self.conf, 
            'rootnode': self.rootnode,
            'generated_dict': self.generated_dict,
            'total_valid_num': self.total_valid_num,
        }
        
        with open(ckpt_path, mode='wb') as f:
            pickle.dump(cp_obj, f)
        self.flush()
            


from abc import ABC, abstractmethod
from typing import Dict

from rdkit.Chem import Mol


class Filter(ABC):
    @staticmethod
    @abstractmethod
    def check(mol: Mol, config: Dict) -> bool:
        raise NotImplementedError('Please check your filter file')



import argparse
import os
import re

import pandas as pd
from rdkit import Chem
import yaml

def get_parser():
    parser = argparse.ArgumentParser(
        description="",
        usage=f"chemtsv2-add_cores_to_linker -c PATH_TO_CONFIG_FILE"
    )
    parser.add_argument(
        "-c", "--config", type=str, required=True,
        help="path to the config file used with the molecule generation"
    )
    return parser.parse_args()

    
def main():
    def __add_atom_index_in_wildcard(smiles: str):
        c = iter(range(1, smiles.count('*')+1))
        labeled_smiles = re.sub(r'\*', lambda _: f'[*:{next(c)}]', smiles)
        return labeled_smiles

    def __build_smiles_from_cores_and_linker(linker, cores):
        if linker.count("*") != len(cores):
            return None
        linker_mol = Chem.MolFromSmiles(__add_atom_index_in_wildcard(linker))
        rwmol = Chem.RWMol(linker_mol)
        for m in cores:
            rwmol.InsertMol(m)
        prod = Chem.molzip(rwmol)
        return Chem.MolToSmiles(prod)
    
    args = get_parser()
    config_file = args.config
    print(f"[INFO] Load config file: {args.config}")
    with open(config_file, 'r') as f:
        conf = yaml.load(f, Loader=yaml.SafeLoader)
    result_path = os.path.join(conf['output_dir'], f"result_C{conf['c_val']}.csv")
    print(f"[INFO] Load result CSV file: {result_path}")
    df = pd.read_csv(result_path)
    cores_mol = [Chem.MolFromSmiles(s) for s in conf['cores']]
    df['smiles_w_cores'] = df['smiles'].apply(lambda x: __build_smiles_from_cores_and_linker(x, cores_mol))

    stem, ext = os.path.splitext(result_path)
    output_fname = f"{stem}_add_cores{ext}"
    df.to_csv(output_fname, mode='w', index=False)
    print(f"[INFO] Save to {output_fname}\n"
          f"[INFO] Done!")


if __name__ == "__main__":
    main()



import argparse
from logging import getLogger, StreamHandler, FileHandler, Formatter, INFO, DEBUG
from importlib import import_module
import os
os.environ['TF_CPP_MIN_LOG_LEVEL']='3'
import pickle
import re
import sys
sys.path.append(os.getcwd())

from mpi4py import MPI
import numpy as np
from numpy.random import default_rng
from rdkit import RDLogger
import tensorflow as tf
tf.compat.v1.disable_eager_execution()
import requests
import yaml

from chemtsv2.utils import loaded_model, get_model_structure_info
from chemtsv2.parallel_mcts import p_mcts


def get_parser():
    parser = argparse.ArgumentParser(
        description="",
        usage=f"python {os.path.basename(__file__)} -c CONFIG_FILE"
    )
    parser.add_argument(
        "-c", "--config", type=str, required=True,
        help="path to a config file"
    )
    parser.add_argument(
        "-d", "--debug", action='store_true',
        help="debug mode"
    )
    parser.add_argument(
        "-g", "--gpu", type=str,
        help="constrain gpu. (e.g. 0,1)"
    )
    parser.add_argument(
        "--use_gpu_only_reward", action='store_true',
        help="use GPUs exclusively for reward calculations"
    )
    return parser.parse_args()
    

def get_logger(level, save_dir, rank):
    logger = getLogger(__name__)
    logger.setLevel(level)
    logger.propagate = False

    formatter = Formatter("%(asctime)s : %(levelname)s : %(message)s ")

    fh = FileHandler(filename=os.path.join(save_dir, f"run_rank{rank}.log"), mode='w')
    fh.setLevel(level)
    fh.setFormatter(formatter)
    sh = StreamHandler()
    sh.setLevel(level)
    sh.setFormatter(formatter)

    logger.addHandler(fh)
    logger.addHandler(sh)
    return logger


def set_default_config(conf):
    conf.setdefault('c_val', 1.0)
    conf.setdefault('hours', 0.02) 
    conf.setdefault('output_dir', 'result/example_mp01')
    conf.setdefault('zobrist_hash_seed', 3)
    conf.setdefault('fix_random_seed', False)
    conf.setdefault('random_seed', -1)
    conf.setdefault('token', 'model/tokens.pkl')

    conf.setdefault('model_setting', {
        'model_json': 'model/model.tf25.json',
        'model_weight': 'model/model.tf25.best.ckpt.h5'})
    conf.setdefault('reward_setting', {
        'reward_module': 'reward.logP_reward',
        'reward_class': 'LogP_reward'})

    conf.setdefault('search_type', 'MP_MCTS')

    conf.setdefault('use_lipinski_filter', False)
    conf.setdefault('lipinski_filter', {
        'module': 'filter.lipinski_filter',
        'class': 'LipinskiFilter',
        'type': 'rule_of_5'})
    conf.setdefault('use_radical_filter', False)
    conf.setdefault('radical_filter', {
        'module': 'filter.radical_filter',
        'class': 'RadicalFilter'})
    conf.setdefault('use_pubchem_filter', False) 
    conf.setdefault('pubchem_filter', {
        'module': 'filter.pubchem_filter',
        'class': 'PubchemFilter'}) 
    conf.setdefault('use_sascore_filter', False)
    conf.setdefault('sascore_filter', {
        'module': 'filter.sascore_filter',
        'class': 'SascoreFilter',
        'threshold': 3.5})
    conf.setdefault('use_ring_size_filter', False)
    conf.setdefault('ring_size_filter', {
        'module': 'filter.ring_size_filter',
        'class': 'RingSizeFilter',
        'threshold': 6})
    conf.setdefault('use_pains_filter', False)
    conf.setdefault('pains_filter', {
        'module': 'filter.pains_filter',
        'class': 'PainsFilter',
        'type': ['pains_a']})
    conf.setdefault('use_selfies', False)

    
def get_filter_modules(conf):
    pat = re.compile(r'^use.*filter$')
    module_list = []
    for k, frag in conf.items():
        if not pat.search(k) or frag != True:
            continue
        _k = k.replace('use_', '')
        module_list.append(getattr(import_module(conf[_k]['module']), conf[_k]['class']))
    return module_list


def main():
    # Initialize MPI environment
    comm = MPI.COMM_WORLD
    rank = comm.Get_rank()
    nprocs = comm.Get_size()
    status = MPI.Status()
    mem = np.zeros(1024 * 10 * 1024)
    MPI.Attach_buffer(mem)
    
    args = get_parser()
    with open(args.config, 'r') as f:
        conf = yaml.load(f, Loader=yaml.SafeLoader)
    set_default_config(conf)
    os.makedirs(conf['output_dir'], exist_ok=True)
    os.environ['CUDA_VISIBLE_DEVICES'] = '-1' if args.gpu is None else args.gpu
    
    conf['debug'] = args.debug
    log_level = DEBUG if args.debug else INFO
    logger = get_logger(log_level, conf["output_dir"], rank)
    if not conf['debug']:
        import warnings
        warnings.filterwarnings('ignore')
        RDLogger.DisableLog("rdApp.*")
    if args.use_gpu_only_reward:
        logger.info("Use GPUs exclusively for reward caluculations")
        tf.config.set_visible_devices([], 'GPU')
    
    if args.debug:
        conf['fix_random_seed'] = True
        conf['random_seed'] = 1234

    if conf['random_seed'] != -1:
        conf['fix_random_seed'] = True

    # download additional data if files don't exist
    if not os.path.exists('data/sascorer.py'):
        url = 'https://raw.githubusercontent.com/rdkit/rdkit/master/Contrib/SA_Score/sascorer.py'
        with open('data/sascorer.py', 'w') as f:
            f.write(requests.get(url).text)
    if not os.path.exists('data/fpscores.pkl.gz'):
        url = 'https://raw.githubusercontent.com/rdkit/rdkit/master/Contrib/SA_Score/fpscores.pkl.gz'
        with open('data/fpscores.pkl.gz', 'wb') as f:
            f.write(requests.get(url).content)

    with open(conf['token'], 'rb') as f:
        tokens = pickle.load(f)
    conf['token'] = tokens
    conf['max_len'], conf['rnn_vocab_size'], conf['rnn_output_size'], conf['num_gru_units'] = get_model_structure_info(conf['model_setting']['model_json'], logger)

    rs = conf['reward_setting']
    reward_calculator = getattr(import_module(rs['reward_module']), rs['reward_class'])

    if rank == 0:
        logger.info(f"========== Configuration ==========")
        for k, v in conf.items():
            logger.info(f"{k}: {v}")
        logger.info(f"GPU devices: {os.environ['CUDA_VISIBLE_DEVICES']}")
        logger.info(f"===================================")

    conf['filter_list'] = get_filter_modules(conf)

    conf['random_generator'] = default_rng(conf['random_seed']) if conf['fix_random_seed'] else default_rng()

    chem_model = loaded_model(conf['model_setting']['model_weight'], logger, conf)

    logger.info(f'Run MPChemTS [rank {rank}]')
    comm.barrier()
    search = p_mcts(comm, chem_model, reward_calculator, conf, logger)
    if conf['search_type'] == 'TDS_UCT':
        search.TDS_UCT()
    #elif conf['search_type'] == 'TDS_df_UCT':
    #    search.TDS_df_UCT()
    elif conf['search_type'] == 'MP_MCTS':
        search.MP_MCTS()
    else:
        logger.error('[ERROR] Select a search type from [TDS_UCT, MP_MCTS]')
        sys.exit(1)

    comm.barrier()

    logger.info(f"Done MCTS execution [rank {rank}]")

    search.gather_results()
    comm.barrier()
    if rank==0:
        search.flush()
        logger.info("FINISH!")
    comm.barrier()
    MPI.Finalize()

if __name__ == "__main__":
    main()



import importlib
from importlib import import_module
import pickle
import sys
import os

sys.path.append('./')

def calc():
    
    calc_obj = sys.argv[1]
    print(os.getcwd(), calc_obj)
    
    with open(calc_obj, mode='rb') as f:
        mol, conf = pickle.load(f)
    print(conf)
    
    rs = conf['reward_setting']
    reward_calculator = getattr(import_module(rs["reward_module"]), rs["reward_class"])
    
    print(reward_calculator)
    values = [f(mol) for f in reward_calculator.get_objective_functions(conf)]
    print('values', values)
    
    gid = conf['gid']
    result_dir = os.path.join(conf['output_dir'], 'gaussian_result')
    calc_dir = f'InputMolopt{gid}'
    
    if not os.path.exists(result_dir):
        os.mkdir(result_dir)
    if not os.path.exists(os.path.join(result_dir, calc_dir)):
        os.mkdir(os.path.join(result_dir, calc_dir))
    
    with open(os.path.join(result_dir, calc_dir, 'values.pickle'), mode='wb') as f:
        pickle.dump(values, f)
    os.remove(calc_obj)

if __name__ == "__main__":
    calc()



import re

import selfies as sf


def tokenize_smiles(smiles_list, use_selfies=False):
    tokenized_smiles_list = []
    unique_token_set = set()
    for smi in smiles_list:
        tokenized_smiles = selfies_tokenizer_from_smiles(smi) if use_selfies else smi_tokenizer(smi)
        tokenized_smiles.append('\n')
        unique_token_set |= set(tokenized_smiles)
        tokenized_smiles_list.append(tokenized_smiles)
    return sorted(list(unique_token_set)), tokenized_smiles_list


def read_smiles_dataset(filepath):
    with open(filepath, 'r') as f:
        smiles_list = [l.strip('\n') for l in f.readlines()]
    return smiles_list


def smi_tokenizer(smi):
    """
    This function is based on https://github.com/pschwllr/MolecularTransformer#pre-processing
    Modified by Shoichi Ishida
    """
    pattern =  "(\[[^\]]+]|Br?|Cl?|N|O|S|P|F|I|b|c|n|o|s|p|\(|\)|\.|=|#|-|\+|\\\\|\/|:|~|@|\?|>|\*|\$|\%[0-9]{2}|[0-9])"
    regex = re.compile(pattern)
    tokens = [token for token in regex.findall(smi)]
    assert smi == ''.join(tokens)
    tokens.insert(0, '&')
    return tokens


def selfies_tokenizer_from_smiles(smi):
    if '[*]' in smi:
        smi = smi.replace("[*]", "[Lr]")  # Because SELFIES (v2.1.0) currently does not support a wildcard (*) representation.
    slfs = sf.encoder(smi)
    tokens = list(sf.split_selfies(slfs))
    assert slfs == ''.join(tokens)
    tokens.insert(0, '&')
    return tokens